#include <errno.h>
#include <fcntl.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <ctype.h>

#include <sys/signal.h>
#include <sys/poll.h>
#include <sys/stat.h>
#include <sys/wait.h>

#include <popt.h>

#include <regex.h>

#include "initlog.h"
#include "process.h"
#include "fds.h"
#include "alloc.h"

static int
forkCommand (char *const args[], int *outfd, int *errfd, int *cmdfd,
	     int quiet)
{
	/* Fork command 'cmd', returning pid, and optionally pointer
	 * to open file descriptor fd */
	int     fdout = -1, fderr = -1, fdcmd;
	int     outpipe[2], errpipe[2], fdpipe[2];
	int     ourpid, pid;

	if ((pipe (outpipe) < 0) || (pipe (errpipe) < 0)
	    || (pipe (fdpipe) < 0))
	{
		error (EXIT_SUCCESS, errno, "pipe");
		return -1;
	}

	if (outfd)
	{
		fdout = outpipe[1];
		*outfd = outpipe[0];
	} else
	{
		if (!quiet)
			fdout = dup (STDOUT_FILENO);
	}
	if (errfd)
	{
		fderr = errpipe[1];
		*errfd = errpipe[0];
	} else
	{
		if (!quiet)
			fderr = dup (STDERR_FILENO);
	}

	if (cmdfd)
	{
		fdcmd = fdpipe[1];
		*cmdfd = fdpipe[0];
	} else
	{
		fdcmd = open (_PATH_DEVNULL, O_WRONLY);
	}

	ourpid = getpid ();
	pid = fork ();
	if (pid < 0)
	{
		error (EXIT_SUCCESS, errno, "fork");
		return -1;
	}

	/* We exec the command normally as the child. However, if we're getting passed
	 * back arguments via an fd, we'll exec it as the parent. Therefore, if Bill
	 * fucks up and we segfault or something, we don't kill rc.sysinit. */
	if ((cmdfd && !pid) || (pid && !cmdfd))
	{
		/* parent */
		close (fdout);
		close (fderr);
		close (fdcmd);
		if (!pid)
			return ourpid;
		else
			return pid;
	} else
	{
		/* child */
		if (outfd)
		{
			if (dup2 (fdout, STDOUT_FILENO) < 0)
				error (-1, errno, "dup2");
		} else if (quiet)
		{
			int     fd = open (_PATH_DEVNULL, O_WRONLY);

			if (fd < 0
			    || (fd != STDOUT_FILENO
				&& dup2 (fd, STDOUT_FILENO) < 0)
			    || close (fd))
				error (-1, errno, "dup2");
		}

		if (errfd)
		{
			if (dup2 (fderr, STDERR_FILENO) < 0)
				error (-1, errno, "dup2");
		} else if (quiet)
		{
			int     fd = open (_PATH_DEVNULL, O_WRONLY);

			if (fd < 0
			    || (fd != STDERR_FILENO
				&& dup2 (fd, STDERR_FILENO) < 0)
			    || close (fd))
				error (-1, errno, "dup2");
		}

		if (dup2 (fdcmd, CMD_FD) < 0)
			error (-1, errno, "dup2");

		close (fdout);
		close (fderr);
		close (fdcmd);
		if (outfd)
			close (*outfd);
		if (errfd)
			close (*errfd);
		if (cmdfd)
			close (*cmdfd);

		{
			/* close up extra fds, and hope this doesn't break anything */
			int     sc_open_max = sysconf (_SC_OPEN_MAX);

			if (sc_open_max > 1)
			{
				int     fd;

				for (fd = 3; fd < sc_open_max; ++fd)
				{
					if (!cmdfd || fd != CMD_FD)
						close (fd);
				}
			}
		}

		execvp (args[0], args);
		error (-1, errno, "execvp");
		exit (-1);
	}
}

static int monitor( const char *cmdname, int pid, unsigned numfds, int *fds, int reexec, int quiet, int debug )
{
	struct pollfd *pfds = xmalloc( numfds * sizeof(struct pollfd) );
	char *outbuf = 0;
	char *procpath = 0;
	unsigned i;
	int rc = -1;
	int done = 0;

	if ( reexec )
	{
		const char prefix[] = "/proc/";
		unsigned size = sizeof(prefix) + 3 * sizeof( unsigned );
		procpath = alloca( size );
		snprintf( procpath, size, "%s%u", prefix, (unsigned)pid );
	}

	for ( i=0; i<numfds; ++i )
	{
		pfds[i].fd = fds[i];
		pfds[i].events = POLLIN | POLLPRI;
	}

	while ( !done )
	{
		if ( poll( pfds, numfds, 500 ) < 0 )
		{
			if ( EINTR == errno )
				continue;
			error( EXIT_SUCCESS, errno, "poll" );
			return -1;
		}
		if ( !reexec )
		{
			int st = waitpid( pid, &rc, WNOHANG );
			if ( st < 0 )
			{
				error( EXIT_SUCCESS, errno, "waitpid" );
				return -1;
			}
			if ( st )
				done = 1;
		} else
		{
			struct stat sbuf;
			/* if /proc/pid ain't there and /proc is, it's dead... */
			if ( stat( procpath, &sbuf ) && !stat( "/proc/cmdline", &sbuf ) )
				done = 1;
		}

		for( i=0; i < numfds; ++i )
		{
			if ( (pfds[i].revents & (POLLIN | POLLPRI)) )
			{
				unsigned bufsize = 8192 + 1;
				int bytesread = 0;

				do
				{
					char *buf = xmalloc( bufsize );
					bytesread = read( pfds[i].fd, buf, bufsize - 1 );
					if ( bytesread < 0 )
					{
						error( EXIT_SUCCESS, errno, "read" );
						return -1;
					}
					if ( bytesread )
					{
						char *tmpstr = 0;

						if ( !quiet && !reexec )
							write_loop(STDOUT_FILENO, buf, bytesread);

						if ( quiet )
						{
							buf[bytesread] = '\0';
							if ( outbuf )
							{
								outbuf = xrealloc( outbuf, strlen(outbuf) + bytesread + 1 );
								strcat( outbuf, buf );
							}
							else
								outbuf = xstrdup( buf );
						}
						while ( (tmpstr = getLine( &buf )) )
						{
							int ignore=0;

							if ( regList )
							{
								unsigned j;
								for ( j = 0; regList[j]; ++j )
									if ( !regexec( regList[j], tmpstr, 0, NULL, 0 ) )
									{
										ignore = 1;
										break;
									}
							}
							if ( !ignore )
							{
								if ( !reexec )
								{
									if ( getenv("IN_INITLOG") )
									{
										char buffer[8192];
										DDEBUG( "sending =%s= to initlog parent\n", tmpstr );
										snprintf( buffer, sizeof buffer, "-n %s -s \"%s\"\n",
												cmdname, tmpstr );
										/* don't blow up if parent isn't there */
										signal( SIGPIPE, SIG_IGN );
										write_loop(CMD_FD, buffer, strlen(buffer));
										signal( SIGPIPE, SIG_DFL );
									} else
										logString( cmdname, tmpstr );
								} else {
									const char **cmdargs = 0;
									const char **tmpargs = 0;
									int cmdargc = 0, z;

									poptParseArgvString( tmpstr, &cmdargc, &tmpargs );
									cmdargs = xmalloc( (cmdargc+2) * sizeof(char *) );
									for ( z=0; z<cmdargc; ++z )
										cmdargs[z+1] = tmpargs[z];
									cmdargs[0] = xstrdup( "initlog" );
									cmdargs[cmdargc+1] = 0;
									processArgs( cmdargc+1, cmdargs, 1 );
								}
							}
						}
					}
				} while ( bytesread >= bufsize );
			}
		}
	}
	if ( (!WIFEXITED(rc)) || (rc = WEXITSTATUS(rc)) )
	{
		/* If there was an error and we're quiet, be loud */
		if ( quiet && outbuf )
			write_loop(STDOUT_FILENO, outbuf, strlen(outbuf));
		return rc;
	}
	return 0;
}

int
runCommand (const char *cmdname, const char *cmd, int reexec, int quiet,
	    int debug)
{
	int     fds[2];
	int     i, x;
	char  **args;
	const char **tmpargs;

	poptParseArgvString (cmd, &x, &tmpargs);
	args = xmalloc ((x + 1) * sizeof (char *));
	for (i = 0; i < x; ++i)
		args[i] = xstrdup (tmpargs[i]);
	args[i] = 0;
	if (!cmdname || !*cmdname)
	{
		if (strcmp (args[0], "sh") && strcmp (args[0], "/bin/sh"))
			cmdname = basename (args[0]);
		else
			cmdname = basename (args[1]);

		if ((cmdname[0] == 'K' || cmdname[0] == 'S')
		    && isdigit (cmdname[1]) && isdigit (cmdname[2]))
			cmdname += 3;
	}
	if (!reexec)
	{
		int     pid =
			forkCommand (args, &fds[0], &fds[1], NULL, quiet);
		x = monitor (cmdname, pid, 2, fds, reexec, quiet, debug);
	} else
	{
		int     pid;

		setenv ("IN_INITLOG", "yes", 1);
		pid = forkCommand (args, NULL, NULL, &fds[0], quiet);
		unsetenv ("IN_INITLOG");
		x = monitor (cmdname, pid, 1, &fds[0], reexec, quiet, debug);
	}
	return x;
}
